/* Copyright (C) InMon Corporation 2001 ALL RIGHTS RESERVED */
/* $Header: /usr/src/redhat/SOURCES/sflow_tools-1.3/RCS/sflowtool.c,v 1.15 2002/02/07 20:46:39 nhm Exp $ */

#include <config.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>

#ifdef WIN32
#include "winsock2.h"
#else
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#endif

#include "sflowtool.h"

#define YES 1
#define NO 0

typedef struct _SFConfig {
  /* sflow options */
  u_int16_t sFlowInputPort;
  /* netflow options */
  u_int16_t netFlowOutputPort;
  struct in_addr netFlowOutputIP;
  int netFlowOutputSocket;
  u_int16_t netFlowPeerAS;
  int netFlowScale;
  /* tcpdump options */
  int tcpdumpFormat;
  u_int32_t tcpdumpHdrPad;
  u_char zeroPad[100];
} SFConfig;

/* make the options structure global to the program */
static SFConfig sfConfig;

/* define my own IP header struct - to ease portability */
struct myiphdr
  {
    u_int8_t version_and_headerLen;
    u_int8_t tos;
    u_int16_t tot_len;
    u_int16_t id;
    u_int16_t frag_off;
    u_int8_t ttl;
    u_int8_t protocol;
    u_int16_t check;
    u_int32_t saddr;
    u_int32_t daddr;
};

/* same for tcp */
struct mytcphdr
  {
    u_int16_t th_sport;		/* source port */
    u_int16_t th_dport;		/* destination port */
    u_int32_t th_seq;		/* sequence number */
    u_int32_t th_ack;		/* acknowledgement number */
    u_int8_t th_off_and_unused;
    u_int8_t th_flags;
    u_int16_t th_win;		/* window */
    u_int16_t th_sum;		/* checksum */
    u_int16_t th_urp;		/* urgent pointer */
};

/* and UDP */
struct myudphdr {
  u_int16_t uh_sport;           /* source port */
  u_int16_t uh_dport;           /* destination port */
  u_int16_t uh_ulen;            /* udp length */
  u_int16_t uh_sum;             /* udp checksum */
};

/* and ICMP */
struct myicmphdr
{
  u_int8_t type;		/* message type */
  u_int8_t code;		/* type sub-code */
  /* ignore the rest */
};

typedef struct _SFSample {
  struct in_addr sourceIP;
  struct in_addr agent_addr;

  /* the raw pdu */
  u_char *rawSample;
  u_int rawSampleLen;

  u_int sampleType;
  u_int samplerId;

  /* interface info */
  u_long ifIndex;
  u_long networkType;
  u_int64_t ifSpeed;
  u_long ifDirection;
  u_long ifStatus;

  /* sample stream info */
  u_long sysUpTime;
  u_long sequenceNo;
  u_long sampledPacketSize;
  u_long samplesGenerated;
  u_long meanSkipCount;
  u_long samplePool;
  u_long dropEvents;

  /* the sampled header */
  u_long packet_data_tag;
  u_long headerProtocol;
  u_char *header;
  int headerLen;
  
  /* header decode */
  int offsetToIPV4;
  struct in_addr dcd_srcIP;
  struct in_addr dcd_dstIP;
  u_int dcd_ipProtocol;
  u_int dcd_ipTos;
  u_int dcd_ipTTL;
  u_int dcd_sport;
  u_int dcd_dport;
  u_int dcd_tcpFlags;

  /* ports */
  u_long inputPort;
  u_long outputPort;

  /* vlan */
  u_long in_vlan;
  u_long in_priority;
  u_long internalPriority;
  u_long out_vlan;
  u_long out_priority;

  /* extended data fields */
  u_long num_extended;
  u_long extended_data_tag;
#define SASAMPLE_EXTENDED_DATA_SWITCH 1
#define SASAMPLE_EXTENDED_DATA_ROUTER 4
#define SASAMPLE_EXTENDED_DATA_GATEWAY 8
#define SASAMPLE_EXTENDED_DATA_USER 16

  /* IP forwarding info */
  struct in_addr nextHop;
  u_long srcMask;
  u_long dstMask;
  u_long my_as;
  u_long src_as;
  u_long src_peer_as;
  u_long dst_as_path_len;
  u_long *dst_as_path;

  /* user id */
#define SA_MAX_EXTENDED_USER_LEN 200
  u_int src_user_len;
  char src_user[SA_MAX_EXTENDED_USER_LEN+1];
  u_int dst_user_len;
  char dst_user[SA_MAX_EXTENDED_USER_LEN+1];

  /* counter blocks */
  u_long statsSamplingInterval;
  u_long counterBlockVersion;

} SFSample;

/* tcpdump file format */

struct pcap_file_header {
  u_int32_t magic;
  u_int16_t version_major;
  u_int16_t version_minor;
  u_int32_t thiszone;	/* gmt to local correction */
  u_int32_t sigfigs;	/* accuracy of timestamps */
  u_int32_t snaplen;	/* max length saved portion of each pkt */
  u_int32_t linktype;	/* data link type (DLT_*) */
};

struct pcap_pkthdr {
  struct timeval ts;	/* time stamp */
  u_int32_t caplen;	/* length of portion present */
  u_int32_t len;	/* length this packet (off wire) */
  /* some systems expect to see more information here. For example,
   * on some versions of RedHat Linux, there are three extra fields:
   *   int index;
   *   unsigned short protocol;
   *   unsigned char pkt_type;
   * To pad the header with zeros, use the tcpdumpHdrPad option.
   */
};

/* netflow version 5 file format */

typedef struct _NFFlow5 {
  u_int32_t srcIP;
  u_int32_t dstIP;
  u_int32_t nextHop;
  u_int16_t if_in;
  u_int16_t if_out;
  u_int32_t frames;
  u_int32_t bytes;
  u_int32_t firstTime;
  u_int32_t lastTime;
  u_int16_t srcPort;
  u_int16_t dstPort;
  u_int8_t pad1;
  u_int8_t tcpFlags;
  u_int8_t ipProto;
  u_int8_t ipTos;
  u_int16_t srcAS;
  u_int16_t dstAS;
  u_int8_t srcMask;  /* No. bits */
  u_int8_t dstMask;  /* No. bits */
  u_int16_t pad2;
} NFFlow5;

typedef struct _NFFlowHdr5 {
  u_int16_t version;
  u_int16_t count;
  u_int32_t sysUpTime;
  u_int32_t unixSeconds;
  u_int32_t unixNanoSeconds;
  u_int32_t flowSequence;
  u_int8_t engineType;
  u_int8_t engineId;
  u_int16_t sampling_interval;
} NFFlowHdr5;

typedef struct _NFFlowPkt5 {
  NFFlowHdr5 hdr;
  NFFlow5 flow; /* normally an array, but here we always send just 1 at a time */
} NFFlowPkt5;


/*_________________---------------------------__________________
  _________________        sf_log             __________________
  -----------------___________________________------------------
*/

void sf_log(char *fmt, ...)
{
  /* don't print anything if we are exporting tcpdump format instead */
  if(sfConfig.tcpdumpFormat == NO) {
    va_list args;
    va_start(args, fmt);
    vprintf(fmt, args);
  }
}

/*_________________---------------------------__________________
  _________________        printHex           __________________
  -----------------___________________________------------------
*/

static u_char bin2hex(int nib) { return (nib < 10) ? ('0' + nib) : ('A' - 10 + nib); }

int printHex(const u_char *a, int len, u_char *buf, int bufLen, int marker, int bytesPerOutputLine)
{
  int b = 0, i = 0;
  for(; i < len; i++) {
    u_char byte;
    if(b > (bufLen - 10)) break;
    if(marker > 0 && i == marker) {
      buf[b++] = '<';
      buf[b++] = '*';
      buf[b++] = '>';
      buf[b++] = '-';
    }
    byte = a[i];
    buf[b++] = bin2hex(byte >> 4);
    buf[b++] = bin2hex(byte & 0x0f);
    if(i > 0 && (i % bytesPerOutputLine) == 0) buf[b++] = '\n';
    else {
      // separate the bytes with a dash
      if (i < (len - 1)) buf[b++] = '-';
    }
  }
  buf[b] = '\0';
  return b;
}

/*_________________---------------------------__________________
  _________________      IP_to_a              __________________
  -----------------___________________________------------------
*/

char *IP_to_a(u_long ipaddr, char *buf)
{
  u_char *ip = (u_char *)&ipaddr;
  sprintf(buf, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
  return buf;
}

/*_________________---------------------------__________________
  _________________    receiveError           __________________
  -----------------___________________________------------------
*/

static void receiveError(SFSample *sample, char *errm, int hexdump, u_char *currentMark)
{
  char ipbuf[51];
  u_char scratch[6000];
  char *msg = "";
  char *hex = "";
  u_long markOffset = 0;
  if(currentMark != NULL) markOffset = currentMark - sample->rawSample;
  if(errm) msg = errm;
  if(hexdump) {
    printHex(sample->rawSample, sample->rawSampleLen, scratch, 6000, markOffset, 16);
    hex = scratch;
  }
  fprintf(stderr, "%s (source IP = %s) %s\n", msg, IP_to_a(sample->sourceIP.s_addr, ipbuf), hex);
}

/*_________________---------------------------__________________
  _________________     decodeLinkLayer       __________________
  -----------------___________________________------------------
  store the offset to the start of the ipv4 header in the sequence_number field
  or -1 if not found. Decode the 802.1d if it's there.
*/

#define NFT_ETHHDR_SIZ 14
#define NFT_8022_SIZ 3
#define NFT_MAX_8023_LEN 1500

#define NFT_MIN_SIZ (NFT_ETHHDR_SIZ + sizeof(struct myiphdr))

static void decodeLinkLayer(SFSample *sample)
{
  u_char *start = (u_char *)sample->header;
  u_char *end = start + sample->headerLen;
  u_char *ptr = start;
  u_int16_t type_len;

  /* assume not found */
  sample->offsetToIPV4 = -1;

  if(sample->headerLen < NFT_ETHHDR_SIZ) return; /* not enough for an Ethernet header */

  sf_log("dstMAC %02x%02x%02x%02x%02x%02x\n", ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5]);
  ptr += 6;
  sf_log("srcMAC %02x%02x%02x%02x%02x%02x\n", ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5]);
  ptr += 6;
  type_len = (ptr[0] << 8) + ptr[1];
  ptr += 2;

  if(type_len == 0x8100) {
    /* VLAN  - next two bytes */
    u_int32_t vlanData = (ptr[0] << 8) + ptr[1];
    u_int32_t vlan = vlanData & 0x0fff;
    u_int32_t priority = vlanData >> 13;
    /*  _____________________________________ */
    /* |   pri  | c |         vlan-id        | */
    /*  ------------------------------------- */
    /* [priority = 3bits] [Canonical Format Flag = 1bit] [vlan-id = 12 bits] */
    sf_log("decodedVLAN %lu\n", vlan);
    sf_log("decodedPriority %lu\n", priority);
    /* now get the type_len again (next two bytes) */
    type_len = (ptr[0] << 8) + ptr[1];
  }

  /* now we're just looking for IP */
  if(sample->headerLen < NFT_MIN_SIZ) return; /* not enough for an IPv4 header */
  
  /* peek for IPX */
  if(type_len == 0x0200 || type_len == 0x0201 || type_len == 0x0600) {
#define IPX_HDR_LEN 30
#define IPX_MAX_DATA 546
    int ipxChecksum = (ptr[0] == 0xff && ptr[1] == 0xff);
    int ipxLen = (ptr[2] << 8) + ptr[3];
    if(ipxChecksum &&
       ipxLen >= IPX_HDR_LEN &&
       ipxLen <= (IPX_HDR_LEN + IPX_MAX_DATA))
      /* we don't do anything with IPX here */
      return;
  } 
  
  if(type_len <= NFT_MAX_8023_LEN) {
    /* assume 802.3+802.2 header */
    /* check for SNAP */
    if(ptr[0] == 0xAA &&
       ptr[1] == 0xAA &&
       ptr[2] == 0x03) {
      type_len = (ptr[3] << 8) + ptr[4];
      ptr += 5;
    }
    else {
      if (ptr[0] == 0x06 &&
	  ptr[1] == 0x06 &&
	  (ptr[2] & 0x01)) {
	/* IP over 8022 */
	ptr += 3;
	/* force the type_len to be IP so we can inline the IP decode below */
	type_len = 0x0800;
      }
      else return;
    }
  }
  
  /* assume type_len is an ethernet-type now */
  
  if(type_len == 0x0800) {
    /* IPV4 */
    if((end - ptr) < sizeof(struct myiphdr)) return;
    /* look at first byte of header.... */
    /*  ___________________________ */
    /* |   version   |    hdrlen   | */
    /*  --------------------------- */
    if((*ptr >> 4) != 4) return; /* not version 4 */
    if((*ptr & 15) < 5) return; /* not IP (hdr len must be 5 quads or more) */
    /* survived all the tests - store the offset to the start of the ip header */
    sample->offsetToIPV4 = (ptr - start);
  }
}


/*_________________---------------------------__________________
  _________________     decodeIPV4            __________________
  -----------------___________________________------------------
*/

static void decodeIPV4(SFSample *sample)
{
  if(sample->offsetToIPV4 > 0) {
    char buf[51];
    u_char *ptr = sample->header + sample->offsetToIPV4;
    /* Create a local copy of the IP header (cannot overlay structure in case it is not quad-aligned...some
       platforms would core-dump if we tried that).  It's OK coz this probably performs just as well anyway. */
    struct myiphdr ip;
    memcpy(&ip, ptr, sizeof(ip));
    /* Value copy all ip elements into sample */
    sample->dcd_srcIP.s_addr = ip.saddr;
    sample->dcd_dstIP.s_addr = ip.daddr;
    sample->dcd_ipProtocol = ip.protocol;
    sample->dcd_ipTos = ip.tos;
    sample->dcd_ipTTL = ip.ttl;
    /* Log out the decoded IP fields */
    sf_log("srcIP %s\n", IP_to_a(sample->dcd_srcIP.s_addr, buf));
    sf_log("dstIP %s\n", IP_to_a(sample->dcd_dstIP.s_addr, buf));
    sf_log("IPProtocol %u\n", sample->dcd_ipProtocol);
    sf_log("IPTOS %u\n", sample->dcd_ipTos);
    sf_log("IPTTL %u\n", sample->dcd_ipTTL);
    /* advance the pointer to the next protocol layer */
    ptr += sizeof(struct myiphdr);
    
    switch(ip.protocol) {
    case 1: /* ICMP */
      {
	struct myicmphdr icmp;
	memcpy(&icmp, ptr, sizeof(icmp));
	sf_log("ICMPType %u\n", icmp.type);
	sf_log("ICMPCode %u\n", icmp.code);
      }
      break;
    case 6: /* TCP */
      {
	struct mytcphdr tcp;
	memcpy(&tcp, ptr, sizeof(tcp));
	sample->dcd_sport = ntohs(tcp.th_sport);
	sample->dcd_dport = ntohs(tcp.th_dport);
	sample->dcd_tcpFlags = tcp.th_flags;
	sf_log("TCPSrcPort %u\n", sample->dcd_sport);
	sf_log("TCPDstPort %u\n",sample->dcd_dport);
	sf_log("TCPFlags %u\n", sample->dcd_tcpFlags);
      }
      break;
    case 17: /* UDP */
      {
	struct myudphdr udp;
	memcpy(&udp, ptr, sizeof(udp));
	sample->dcd_sport = ntohs(udp.uh_sport);
	sample->dcd_dport = ntohs(udp.uh_dport);
	sf_log("UDPSrcPort %u\n", sample->dcd_sport);
	sf_log("UDPDstPort %u\n", sample->dcd_dport);
      }
      break;
    default: /* some other protcol */
      break;
    }
  }
}

/*_________________---------------------------__________________
  _________________   writePcapHeader         __________________
  -----------------___________________________------------------
*/

#define TCPDUMP_MAGIC 0xa1b2c3d4  /* from libpcap-0.5: savefile.c */
#define DLT_EN10MB	1	  /* from libpcap-0.5: net/bpf.h */
#define PCAP_VERSION_MAJOR 2      /* from libpcap-0.5: pcap.h */
#define PCAP_VERSION_MINOR 4      /* from libpcap-0.5: pcap.h */

static void writePcapHeader() {
  struct pcap_file_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  hdr.magic = TCPDUMP_MAGIC;
  hdr.version_major = PCAP_VERSION_MAJOR;
  hdr.version_minor = PCAP_VERSION_MINOR;
  hdr.thiszone = 0;
  hdr.snaplen = 128;
  hdr.sigfigs = 0;
  hdr.linktype = DLT_EN10MB;
  if (fwrite((char *)&hdr, sizeof(hdr), 1, stdout) != 1) {
    fprintf(stderr, "failed to write tcpdump header: %s\n", strerror(errno));
    exit(-1);
  }
  fflush(stdout);
}

/*_________________---------------------------__________________
  _________________   writePcapPacket         __________________
  -----------------___________________________------------------
*/

static void writePcapPacket(SFSample *sample) {
  struct pcap_pkthdr hdr;
  hdr.ts.tv_sec = time(NULL);
  hdr.ts.tv_usec = 0;
  hdr.caplen = sample->headerLen;
  hdr.len = sample->sampledPacketSize;

  if(fwrite((char *)&hdr, sizeof(hdr), 1, stdout) != 1) {
    fprintf(stderr, "writePcapPacket: hdr write failed: %s\n", strerror(errno));
    exit(-2);
  }
  if(sfConfig.tcpdumpHdrPad > 0) {
    if(fwrite(sfConfig.zeroPad, sfConfig.tcpdumpHdrPad, 1, stdout) != 1) {
      fprintf(stderr, "writePcapPacket: hdr padding write failed: %s\n", strerror(errno));
      exit(-10);
    }
  }
  if(fwrite((char *)sample->header, sample->headerLen, 1, stdout) != 1) {
    fprintf(stderr, "writePcapPacket: packet write failed: %s\n", strerror(errno));
    exit(-3);
  }
  fflush(stdout);
}

/*_________________---------------------------__________________
  _________________   openNetFlowSocket       __________________
  -----------------___________________________------------------
*/

static void openNetFlowSocket()
{
  struct sockaddr_in addr;
  memset((char *)&addr,0,sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = ntohs(sfConfig.netFlowOutputPort);
  addr.sin_addr.s_addr = sfConfig.netFlowOutputIP.s_addr; 
  if((sfConfig.netFlowOutputSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
    fprintf(stderr, "netflow output socket open failed");
    exit(-4);
  }
  /* connect to it so we can just use send() or write() to send on it */
  if(connect(sfConfig.netFlowOutputSocket, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
    fprintf(stderr, "connect() to netflow output socket failed");
    exit(-5);
  }
}

/*_________________---------------------------__________________
  _________________   sendNetFlowDatagram     __________________
  -----------------___________________________------------------
*/

static NFFlowSequenceNo = 0;

static void sendNetFlowDatagram(SFSample *sample)
{
  NFFlowPkt5 pkt;
  time_t now = time(NULL);
  memset(&pkt, 0, sizeof(pkt));
  pkt.hdr.version = htons(5);
  pkt.hdr.count = htons(1);
  pkt.hdr.sysUpTime = htonl(now % (3600 * 24)) * 1000;  /* pretend we started at midnight (milliseconds) */
  pkt.hdr.unixSeconds = htonl(now);
  pkt.hdr.unixNanoSeconds = 0; /* no need to be more accurate than 1 second */
  pkt.hdr.flowSequence = htonl(NFFlowSequenceNo++);

  pkt.flow.srcIP = sample->dcd_srcIP.s_addr;
  pkt.flow.dstIP = sample->dcd_dstIP.s_addr;
  pkt.flow.nextHop = sample->nextHop.s_addr;
  pkt.flow.if_in = htons((u_int16_t)sample->inputPort);
  pkt.flow.if_out= htons((u_int16_t)sample->outputPort);

  if(sfConfig.netFlowScale) {
    pkt.flow.frames = htonl(sample->meanSkipCount);
    pkt.flow.bytes = htonl(sample->meanSkipCount * sample->sampledPacketSize);
  }
  else {
    /* set the sampling_interval header field too (used to be a 16-bit reserved field) */
    u_int16_t samp_ival = (u_int16_t)sample->meanSkipCount;
    pkt.hdr.sampling_interval = htons(samp_ival & 0x4000);
    pkt.flow.frames = htonl(1);
    pkt.flow.bytes = htonl(sample->sampledPacketSize);
  }
    
  pkt.flow.firstTime = pkt.hdr.sysUpTime;  /* set the start and end time to be now (in milliseconds since last boot) */
  pkt.flow.lastTime =  pkt.hdr.sysUpTime;
  pkt.flow.srcPort = htons((u_int16_t)sample->dcd_sport);
  pkt.flow.dstPort = htons((u_int16_t)sample->dcd_dport);
  pkt.flow.tcpFlags = sample->dcd_tcpFlags;
  pkt.flow.ipProto = sample->dcd_ipProtocol;
  pkt.flow.ipTos = sample->dcd_ipTos;

  if(sfConfig.netFlowPeerAS) {
    pkt.flow.srcAS = htons((u_int16_t)sample->src_peer_as);   
    if(sample->dst_as_path) {
      u_int32_t as32 = ntohl(sample->dst_as_path[0]);
      u_int16_t as16 = (u_int16_t)as32;
      pkt.flow.dstAS = htons(as16);
    }
  }
  else {
    pkt.flow.srcAS = htons((u_int16_t)sample->src_as);   
    if(sample->dst_as_path) {
      u_int32_t as32 = ntohl(sample->dst_as_path[sample->dst_as_path_len - 1]);
      u_int16_t as16 = (u_int16_t)as32;
      pkt.flow.dstAS = htons(as16);
    }
  }
  
  pkt.flow.srcMask = (u_int8_t)sample->srcMask;
  pkt.flow.dstMask = (u_int8_t)sample->dstMask;

  /* send non-blocking */
  send(sfConfig.netFlowOutputSocket, (char *)&pkt, sizeof(pkt), 0);
  
}
      
/*_________________---------------------------__________________
  _________________   read data macros        __________________
  -----------------___________________________------------------
*/

#define GETDATA32(target, datap) (target) = ntohl(*(datap)++)
#define GETDATA32_NOBSWAP(target, datap) (target) = *(datap)++
#define GETDATA64(target, datap) \
  do { u_int64_t tmpLo, tmpHi;   \
       GETDATA32(tmpHi, (datap));  \
       GETDATA32(tmpLo, (datap));  \
       (target) = (tmpHi << 32) + tmpLo; \
  } while(0)


/*_________________---------------------------__________________
  _________________    readExtendedSwitch     __________________
  -----------------___________________________------------------
*/

static u_long *readExtendedSwitch(SFSample *sample, u_long *datap, u_char *endPtr)
{
  GETDATA32(sample->in_vlan, datap);
  GETDATA32(sample->in_priority, datap);
  GETDATA32(sample->out_vlan, datap);
  GETDATA32(sample->out_priority, datap);

  sample->extended_data_tag |= SASAMPLE_EXTENDED_DATA_SWITCH;
  
  sf_log("in_vlan %lu\n", sample->in_vlan);
  sf_log("in_priority %lu\n", sample->in_priority);
  sf_log("out_vlan %lu\n", sample->out_vlan);
  sf_log("out_priority %lu\n", sample->out_priority);
  
  return datap;
}

/*_________________---------------------------__________________
  _________________    readExtendedRouter     __________________
  -----------------___________________________------------------
*/

static u_long *readExtendedRouter(SFSample *sample, u_long *datap, u_char *endPtr)
{
  u_int32_t addrType;
  char buf[51];
  GETDATA32(addrType, datap);
  if(addrType == INMADDRESSTYPE_IP_V4) GETDATA32_NOBSWAP(sample->nextHop.s_addr, datap);
  else {
    fprintf(stderr, "nextHop addrType = %d - currently only IPV4 nexthop supported\n", addrType);
    datap += 4; /* skip over the IPV6 address */
    sample->nextHop.s_addr = 0;
  }
  GETDATA32(sample->srcMask, datap);
  GETDATA32(sample->dstMask, datap);

  sample->extended_data_tag |= SASAMPLE_EXTENDED_DATA_ROUTER;

  sf_log("nextHop %s\n", IP_to_a(sample->nextHop.s_addr, buf));
  sf_log("srcSubnetMask %lu\n", sample->srcMask);
  sf_log("dstSubnetMask %lu\n", sample->dstMask);

  return datap;
}

/*_________________---------------------------__________________
  _________________  readExtendedGateway      __________________
  -----------------___________________________------------------
*/

static u_long *readExtendedGateway(SFSample *sample, u_long *datap, u_char *endPtr)
{
  GETDATA32(sample->my_as, datap);  /* shake yo' ass */
  GETDATA32(sample->src_as, datap);
  GETDATA32(sample->src_peer_as, datap);
  GETDATA32(sample->dst_as_path_len, datap);
  /* just point at the dst_as_path array */
  if(sample->dst_as_path_len > 0) sample->dst_as_path = datap;
  /* and skip over it in the input */
  datap += sample->dst_as_path_len;
  if((u_char *)datap > (endPtr + 1)) {
    receiveError(sample, "datap >= (endp + 1)\n", YES, (u_char *)datap);
    return NULL;
  }
  
  sample->extended_data_tag |= SASAMPLE_EXTENDED_DATA_GATEWAY;
  
  sf_log("my_as %lu\n", sample->my_as);
  sf_log("src_as %lu\n", sample->src_as);
  sf_log("src_peer_as %lu\n", sample->src_peer_as);
  sf_log("dst_as_path_len %lu\n", sample->dst_as_path_len);
  if(sample->dst_as_path_len > 0) {
    u_int i = 0;
    for(; i < sample->dst_as_path_len; i++) {
      if(i == 0) sf_log("dst_as_path ");
      else sf_log("-");
      sf_log("%lu", ntohl(sample->dst_as_path[i]));
    }
    sf_log("\n");
  }
  return datap;
}

/*_________________---------------------------__________________
  _________________    readExtendedUser       __________________
  -----------------___________________________------------------
*/

static u_long *readExtendedUser(SFSample *sample, u_long *datap, u_char *endPtr)
{
  GETDATA32(sample->src_user_len, datap);
  if(sample->src_user_len) {
    if(sample->src_user_len > SA_MAX_EXTENDED_USER_LEN) {
      receiveError(sample, "extended_data: src_user_len > MAX\n", YES, (u_char *)datap);
      return NULL;
    }
    memcpy(datap, sample->src_user, sample->src_user_len);
    datap += (sample->src_user_len + 3 / 4);  /* string is padded to quad boundary */
  }
  sample->src_user[sample->src_user_len] = '\0';
  
  /* repeat for dest */
  GETDATA32(sample->dst_user_len, datap);
  if(sample->dst_user_len) {
    if(sample->dst_user_len > SA_MAX_EXTENDED_USER_LEN) {
      receiveError(sample, "extended_data: sample->dst_user_len > MAX\n", YES, (u_char *)datap);
      return NULL;
    }
    memcpy(datap, sample->dst_user, sample->dst_user_len);
    datap += (sample->dst_user_len + 3 / 4);  /* string is padded to quad boundary */
  }
  sample->dst_user[sample->dst_user_len] = '\0';
  
  sample->extended_data_tag |= SASAMPLE_EXTENDED_DATA_USER;
  
  sf_log("src_user %s\n", sample->src_user);
  sf_log("dst_user %s\n", sample->dst_user);
  
  return datap;
}

/*_________________---------------------------__________________
  _________________  receiveSFlowSample       __________________
  -----------------___________________________------------------
*/

static void receiveSFlowSample(SFSample *sample)
{
  u_int numFlowSamples = 0;
  u_int32_t datagramVersion;
  u_int32_t addressType;
  struct in_addr agentIP;
  u_int32_t samplesInPacket;
  struct timeval now;
  u_long *datap = (u_long *)sample->rawSample;

  now.tv_sec = time(NULL);
  now.tv_usec = 0;
  sf_log("startDatagram =================================\n");
  {
    char buf[51];
    sf_log("datagramSourceIP %s\n", IP_to_a(sample->sourceIP.s_addr, buf));
  }
  sf_log("datagramSize %lu\n", sample->rawSampleLen);
  sf_log("unixSecondsUTC %lu\n", now.tv_sec);

  /* check the version */
  GETDATA32(datagramVersion, datap);
  sf_log("datagramVersion %d\n", datagramVersion);
  if(datagramVersion != 2) {
    receiveError(sample,  "unexpected datagram version number: %d\n", YES, (u_char *)datap);
    return;
  }

  /* get the agent address */
  GETDATA32(addressType, datap);
  if(addressType != INMADDRESSTYPE_IP_V4) {
    receiveError(sample, "currently only support INMADDRESSTYPE_IP_V4 as the agent IP address type", YES, (u_char *)datap);
    return;
  }
  GETDATA32_NOBSWAP(agentIP.s_addr, datap);

  GETDATA32(sample->sequenceNo, datap);  /* this is the packet sequence number */
  GETDATA32(sample->sysUpTime, datap);
  GETDATA32(samplesInPacket, datap);

  {
    char buf[51];
    sf_log("agent %s\n", IP_to_a(agentIP.s_addr, buf));
  }
  sf_log("sysUpTime %lu\n", sample->sysUpTime);
  sf_log("packetSequenceNo %lu\n", sample->sequenceNo);
  sf_log("samplesInPacket %lu\n", samplesInPacket);

  { /* now iterate and pull out the flows and counters samples */
    u_int32_t samp = 0;
    u_char *endPtr = (u_char *)sample->rawSample + sample->rawSampleLen;
   
    for(; samp < samplesInPacket; samp++) {
      u_char *startOfSample = (u_char *)datap;
      
      if((u_char *)datap >= endPtr) {
	receiveError(sample, "datap >= endp", YES, (u_char *)datap);
	return;
      }
      
      GETDATA32(sample->sampleType, datap);
      GETDATA32(sample->samplesGenerated, datap);
      GETDATA32(sample->samplerId, datap);
      sf_log("sampleSequenceNo %lu\n", sample->samplesGenerated);
      {
	u_int32_t ds_class = sample->samplerId >> 24;
	u_int32_t ds_index = sample->samplerId & 0x00ffffff;
	sf_log("sourceId %lu:%lu\n", ds_class, ds_index);
      }

      switch(sample->sampleType) {
      case FLOWSAMPLE:
	{
	  sf_log("sampleType FLOWSAMPLE\n");
	  GETDATA32(sample->meanSkipCount, datap);
	  GETDATA32(sample->samplePool, datap);
	  GETDATA32(sample->dropEvents, datap);
	  GETDATA32(sample->inputPort, datap);
	  GETDATA32(sample->outputPort, datap);
	  sf_log("meanSkipCount %lu\n", sample->meanSkipCount);
	  sf_log("samplePool %lu\n", sample->samplePool);
	  sf_log("dropEvents %lu\n", sample->dropEvents);
	  sf_log("inputPort %lu\n", sample->inputPort);
	  sf_log("outputPort %lu\n", sample->outputPort);
	  
	  GETDATA32(sample->packet_data_tag, datap);

	  switch(sample->packet_data_tag) {
	    
	  case INMPACKETTYPE_HEADER:
	    sf_log("packetDataTag INMPACKETTYPE_HEADER\n");
	    GETDATA32(sample->headerProtocol, datap);
	    GETDATA32(sample->sampledPacketSize, datap);
	    GETDATA32(sample->headerLen, datap);
	    sf_log("headerProtocol %lu\n", sample->headerProtocol);
	    sf_log("sampledPacketSize %lu\n", sample->sampledPacketSize);
	    sf_log("headerLen %lu\n", sample->headerLen);

	    sample->header = (u_char *)datap; /* just point at the header */
	    datap += ((sample->headerLen + 3) / 4); /* quad-alignment is required by XDR */
	    if((u_char *)datap >= endPtr) {
	      receiveError(sample, "datap >= endp (headerLen)", YES, (u_char *)datap);
	      return;
	    }
	    {
	      char scratch[2000];
	      printHex(sample->header, sample->headerLen, scratch, 2000, 0, 2000);
	      sf_log("headerBytes %s\n", scratch);
	    }
	    decodeLinkLayer(sample);
	    if(sample->offsetToIPV4 > 0) {
	      // report the size of the original IPPdu (including the IP header)
	      sf_log("IPSize %d\n",  sample->sampledPacketSize - sample->offsetToIPV4);
	      decodeIPV4(sample);
	    }

	    break;
	    
	  case INMPACKETTYPE_IPV4: 
	    sf_log("packetDataTag INMPACKETTYPE_IPV4\n");
	    sample->headerLen = sizeof(INMSampled_ipv4);
	    sample->header = (u_char *)datap; /* just point at the header */
	    datap += (sample->headerLen + 3) / 4; /* quad-alignment is required by XDR */
	    {
	      char buf[51];
	      INMSampled_ipv4 nfKey;
	      memcpy(&nfKey, sample->header, sizeof(nfKey));
	      sample->sampledPacketSize = ntohl(nfKey.length);
	      sf_log("sampledPacketSize %lu\n", sample->sampledPacketSize); 
	      sf_log("IPSize %d\n",  sample->sampledPacketSize);
	      sample->dcd_srcIP = nfKey.src_ip;
	      sample->dcd_dstIP = nfKey.dst_ip;
	      sample->dcd_ipProtocol = ntohl(nfKey.protocol);
	      sample->dcd_ipTos = ntohl(nfKey.tos);
	      sf_log("srcIP %s\n", IP_to_a(sample->dcd_srcIP.s_addr, buf));
	      sf_log("dstIP %s\n", IP_to_a(sample->dcd_dstIP.s_addr, buf));
	      sf_log("IPProtocol %u\n", sample->dcd_ipProtocol);
	      sf_log("IPTOS %u\n", sample->dcd_ipTos);
	      sample->dcd_sport = ntohl(nfKey.src_port);
	      sample->dcd_dport = ntohl(nfKey.dst_port);
	      switch(sample->dcd_ipProtocol) {
	      case 1: /* ICMP */
		sf_log("ICMPType %u\n", sample->dcd_sport);
		/* not sure about the dest port being icmp code
		   - might just be a repeat of the type */
		break;
	      case 6: /* TCP */
		sf_log("TCPSrcPort %u\n", sample->dcd_sport);
		sf_log("TCPDstPort %u\n", sample->dcd_dport);
		sample->dcd_tcpFlags = ntohl(nfKey.tcp_flags);
		sf_log("TCPFlags %u\n", sample->dcd_tcpFlags);
		break;
	      case 17: /* UDP */
		sf_log("UDPSrcPort %u\n", sample->dcd_sport);
		sf_log("UDPDstPort %u\n", sample->dcd_dport);
		break;
	      default: /* some other protcol */
		break;
	      }
	    }
	    break;
	  case INMPACKETTYPE_IPV6: 
	    sf_log("packetDataTag INMPACKETTYPE_IPV6\n");
	    
	    sample->header = (u_char *)datap; /* just point at the header */
	    datap += (sample->headerLen + 3) / 4; /* quad-alignment is required by XDR */
	    {
	      INMSampled_ipv6 nfKey6;
	      memcpy(&nfKey6, sample->header, sizeof(nfKey6));
	      sample->sampledPacketSize = ntohl(nfKey6.length);
	      sf_log("sampledPacketSize %lu\n", sample->sampledPacketSize); 
	    }
	    /* bug: more decode to do here */
	    break;
	    
	  default:
	    receiveError(sample, "unexpected packet_data_tag", YES, (u_char *)datap);
	    return;
	    break;
	  }
	  
	  /* assume no extended data */
	  sample->extended_data_tag = 0;
	  {
	    u_int x;
	    GETDATA32(sample->num_extended, datap);
	    for(x = 0; x < sample->num_extended; x++) {
	      u_int32_t extended_tag;
	      GETDATA32(extended_tag, datap);
	      switch(extended_tag) {
	      case INMEXTENDED_SWITCH:
		sf_log("extendedType SWITCH\n");
		if((datap = readExtendedSwitch(sample, datap, endPtr)) == NULL) return;
		break;
	      case INMEXTENDED_ROUTER:
		sf_log("extendedType ROUTER\n");
		if((datap = readExtendedRouter(sample, datap, endPtr)) == NULL) return;
		break;
	      case INMEXTENDED_GATEWAY:
		sf_log("extendedType GATEWAY\n");
		if((datap = readExtendedGateway(sample, datap, endPtr)) == NULL) return;
		break;
	      case INMEXTENDED_USER:
		sf_log("extendedType USER\n");
		if((datap = readExtendedUser(sample, datap, endPtr)) == NULL) return;
		break;
	      default:
		receiveError(sample, "unrecognized extended data tag", YES, (u_char *)datap);
		return;
	      }
	    }
	  }
	  /* if we are exporting netflow and we have an IPv4 layer, compose the datagram now */
	  if(sfConfig.netFlowOutputSocket && sample->dcd_srcIP.s_addr != 0) sendNetFlowDatagram(sample);
	  /* if we are writing tcpdump format, write the next packet record now */
	  if(sfConfig.tcpdumpFormat) writePcapPacket(sample);
	}
	break;
	
      case COUNTERSSAMPLE:
	{
	  sf_log("sampleType COUNTERSSAMPLE\n");
	  GETDATA32(sample->statsSamplingInterval, datap);
	  sf_log("statsSamplingInterval %lu\n", sample->statsSamplingInterval);
	  /* now find out what sort of counter blocks we have here... */
	  GETDATA32(sample->counterBlockVersion, datap);
	  sf_log("counterBlockVersion %lu\n", sample->counterBlockVersion);
	  
	  /* first see if we should read the generic stats */
	  switch(sample->counterBlockVersion) {
	  case INMCOUNTERSVERSION_GENERIC:
	  case INMCOUNTERSVERSION_ETHERNET:
	  case INMCOUNTERSVERSION_TOKENRING:
	  case INMCOUNTERSVERSION_FDDI:
	  case INMCOUNTERSVERSION_VG:
	  case INMCOUNTERSVERSION_WAN:
	    {
	      u_int64_t cntr64;
	      /* the first part of the generic counters block is really just more info about the interface. */
	      GETDATA32(sample->ifIndex, datap);
	      GETDATA32(sample->networkType, datap);
	      GETDATA64(sample->ifSpeed, datap);
	      GETDATA32(sample->ifDirection, datap);
	      GETDATA32(sample->ifStatus, datap);
	      sf_log("ifIndex %lu\n", sample->ifIndex);
	      sf_log("networkType %lu\n", sample->networkType);
	      sf_log("ifSpeed %lu\n", sample->ifSpeed);
	      sf_log("ifDirection %lu\n", sample->ifDirection);
	      sf_log("ifStatus %lu\n", sample->ifStatus);
	      
	      /* the generic counters always come first */
	      GETDATA64(cntr64, datap); 
	      sf_log("ifInOctets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInUcastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInMulticastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInBroadcastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInDiscards %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInErrors %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifInUnknownProtos %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("ifOutOctets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifOutUcastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifOutMulticastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifOutBroadcastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifOutDiscards %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifOutErrors %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ifPromiscuousMode %Lu\n", cntr64);
	    }
	    break;

	  case INMCOUNTERSVERSION_VLAN:
	    break;

	  default: 
	    receiveError(sample, "unknown stats version", YES, (u_char *)datap);
	    return;
	    break;
	  }
	  
	  /* now see if there are any specific counter blocks to add */
	  switch(sample->counterBlockVersion) {
	  case INMCOUNTERSVERSION_GENERIC:
	    /* nothing more */
	    break;
	  case INMCOUNTERSVERSION_ETHERNET:
	    {
	      u_int32_t cntr32;
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsAlignmentErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsFCSErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsSingleCollisionFrames %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsMultipleCollisionFrames %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsSQETestErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsDeferredTransmissions %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsLateCollisions %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsExcessiveCollisions %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsInternalMacTransmitErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsCarrierSenseErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsFrameTooLongs %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsInternalMacReceiveErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot3StatsSymbolErrors %lu\n", cntr32);
	    }	  
	    break;
	  case INMCOUNTERSVERSION_TOKENRING:
	    {
	      u_int32_t cntr32;
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsLineErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsBurstErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsACErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsAbortTransErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsInternalErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsLostFrameErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsReceiveCongestions %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsFrameCopiedErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsTokenErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsSoftErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsHardErrors %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsSignalLoss %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsTransmitBeacons %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsRecoverys %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsLobeWires %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsRemoves %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsSingles %lu\n", cntr32);
	      GETDATA32(cntr32, datap); 
	      sf_log("dot5StatsFreqErrors %lu\n", cntr32);
	    }
	    break;
	  case INMCOUNTERSVERSION_FDDI:
	    /* nothing more (for the moment) $$$ */
	    break;
	  case INMCOUNTERSVERSION_VG:
	    {
	      u_int64_t cntr64;
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InHighPriorityFrames %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12InHighPriorityOctets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InNormPriorityFrames %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12InNormPriorityOctets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InIPMErrors %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InOversizeFrameErrors %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InDataErrors %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12InNullAddressedFrames %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12OutHighPriorityFrames %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12OutHighPriorityOctets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("dot12TransitionIntoTrainings %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12HCInHighPriorityOctets %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12HCInNormPriorityOctets %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("dot12HCOutHighPriorityOctets %Lu\n", cntr64);
	    }
	    break;
	  case INMCOUNTERSVERSION_WAN:
	    /* nothing more for the moment $$$ */
	    break;
	  case INMCOUNTERSVERSION_VLAN:
	    {
	      u_int64_t cntr64;
	      GETDATA32(sample->in_vlan, datap);
	      sf_log("in_vlan %lu\n", sample->in_vlan);
	      GETDATA64(cntr64, datap); 
	      sf_log("octets %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("ucastPkts %Lu\n", cntr64);
	      GETDATA64(cntr64, datap); 
	      sf_log("multicastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("broadcastPkts %Lu\n", cntr64);
	      GETDATA32(cntr64, datap); 
	      sf_log("discards %Lu\n", cntr64);
	    }
	    break;
	  default:
	    receiveError(sample, "unknown INMCOUNTERSVERSION", YES, (u_char *)datap);
	    return;
	    break;
	  }
	}
	break;
      default:
	receiveError(sample, "unexpected sample type", YES, (u_char *)datap);
	return;
	break;
      }
      // report the size in bytes that this flowSample or counterSample took up in the datagram
      sf_log("%s %d\n",
	     (sample->sampleType == FLOWSAMPLE ? "flowSampleSize" : "countersSampleSize"),
	     (u_char *)datap - startOfSample);
    }
  }
}

/*__________________-----------------------------__________________
   _________________    openInputUDPSocket       __________________
   -----------------_____________________________------------------
*/

static int openInputUDPSocket(u_int16_t port)
{
  int soc;
  struct sockaddr_in myaddr_in;

  /* Create socket */
  memset((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
  myaddr_in.sin_family = AF_INET;
  myaddr_in.sin_addr.s_addr = INADDR_ANY;
  myaddr_in.sin_port = htons(port);

  if ((soc = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
    fprintf(stderr, "socket() failed, %s\n", strerror(errno));
    exit(-6);
  }

  /* Bind the socket */
  if (bind(soc, (struct sockaddr *)&myaddr_in, sizeof(struct sockaddr_in)) == -1) {
    fprintf(stderr, "bind() failed, port = %d : %s\n", port, strerror(errno));
    exit(-7);
  }

  return soc;
}

/*_________________---------------------------__________________
  _________________       readPacket          __________________
  -----------------___________________________------------------
*/

static void readPacket(int soc)
{
  struct sockaddr_in peer;
  int alen;
  u_char buf[2048];
  SFSample sample;

  int cc = recvfrom(soc, buf, 2048, 0, (struct sockaddr *)&peer, &alen);
  if(cc <= 0) {
    fprintf(stderr, "recvfrom() failed, %s\n", strerror(errno));
    return;
  }

  memset(&sample, 0, sizeof(sample));
  sample.rawSample = buf;
  sample.rawSampleLen = cc;
  sample.sourceIP = peer.sin_addr;
  receiveSFlowSample(&sample);
  fflush(stdout);
}

/*_________________---------------------------__________________
  _________________      instructions         __________________
  -----------------___________________________------------------
*/

static void instructions(char *command)
{
  fprintf(stderr,"Usage: %s [-p port]\n", command);
  fprintf(stderr,"\n");
  fprintf(stderr,"tcpdump options:\n");
  fprintf(stderr, "   -t                 -  (output in binary tcpdump(1) format)\n");
  fprintf(stderr, "   -z pad             -  (extend tcpdump pkthdr with this many zeros\n");
  fprintf(stderr, "                          e.g. try -z 8 for tcpdump on Red Hat Linux)\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"NetFlow options:\n");
  fprintf(stderr, "   -c hostname_or_IP  -  (netflow collector host)\n");
  fprintf(stderr, "   -d port            -  (netflow collector UDP port)\n");
  fprintf(stderr, "   -e                 -  (netflow collector peer_as (default = origin_as))\n");
  fprintf(stderr, "   -s                 -  (disable scaling of netflow output by sampling rate)\n");
  exit(1);
}

/*_________________---------------------------__________________
  _________________   process_command_line    __________________
  -----------------___________________________------------------
*/

static void process_command_line(int argc, char *argv[])
{
  int arg = 1, in = 0;
  
  /* set defaults */
  sfConfig.sFlowInputPort = 6343;
  
  /* walk though the args */
  while (arg < argc) {
    if(argv[arg][0] != '-') instructions(*argv);
    in = argv[arg++][1];
    switch(in) {
    case 'p': sfConfig.sFlowInputPort = atoi(argv[arg++]); break;
    case 't': sfConfig.tcpdumpFormat = YES; break;
    case 'z': sfConfig.tcpdumpHdrPad = atoi(argv[arg++]); break;
    case 'c':
      {
	struct hostent *ent = gethostbyname(argv[arg++]);
	if(ent == NULL) {
	  fprintf(stderr, "netflow collector hostname lookup failed\n");
	  exit(-8);
        }
    	else sfConfig.netFlowOutputIP.s_addr = ((struct in_addr *)(ent->h_addr))->s_addr;
      }
      break;
    case 'd': sfConfig.netFlowOutputPort = atoi(argv[arg++]); break;
    case 'e': sfConfig.netFlowPeerAS = YES; break;
    case 's': sfConfig.netFlowScale = NO; break;
    case '?':
    case 'h':
    default: instructions(*argv);
    }
  }
}

/*_________________---------------------------__________________
  _________________         main              __________________
  -----------------___________________________------------------
*/

int main(int argc, char *argv[])
{
  u_int32_t soc;
  fd_set readfds;

#ifdef WIN32
  WSADATA wsadata;
  WSAStartup(0xffff, &wsadata);
  /* bug: supposed to call WSACleanup() on termination */
#endif

  /* read the command line */
  process_command_line(argc, argv);
  /* open the input socket */
  soc = openInputUDPSocket(sfConfig.sFlowInputPort);
  /* possible open an output socket for netflow */
  if(sfConfig.netFlowOutputPort != 0 && sfConfig.netFlowOutputIP.s_addr != 0) openNetFlowSocket();
  /* if tcpdump format, write the header */
  if(sfConfig.tcpdumpFormat) writePcapHeader();
  /* set the select mask */
  FD_ZERO(&readfds);
  /* loop reading packets */
  for(;;) {
    int nfds;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000;
    FD_SET(soc, &readfds);
    
    nfds = select(soc + 1,
		  &readfds,
		  (fd_set *)NULL,
		  (fd_set *)NULL,
		  &timeout);
    /* we may return prematurely if a signal was caught, in which case
     * nfds will be -1 and errno will be set to EINTR.  If we get any other
     * error, abort.
     */
    if(nfds < 0 && errno != EINTR) {
      fprintf(stderr, "select() returned %d\n", nfds);
      exit(-9);
    }
    if ((nfds > 0) && FD_ISSET(soc, &readfds)) readPacket(soc);
  }
  return 0;
}

